// Copyright (C) 1999-2001 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of ccscript.
// 
// The exception is that, if you link the ccscript library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the ccscript library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name ccscript.  If you copy code from other releases into a copy of
// ccscript, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for ccscript, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include <cc++/config.h>
#include <cc++/macros.h>
#include <cc++/misc.h>
#include <cc++/slog.h>
#include <cc++/file.h>
#include <cstdlib>
#include <fstream>
#include <ctype.h>
#include <cstdio>
#include "script.h"

using namespace std;

static int keyindex(const char *keyword)
{
	unsigned key = 0;

	while(*keyword)
		key ^= key << 1 ^ (*(keyword++) & 0x1f);

	return key % KEYWORD_INDEX_SIZE;
}

ScriptCommand::ScriptCommand(const char *cfgfile) :
Keydata(cfgfile), Mutex()
{
	char *scrname, *ext;
	char buffer[256];
	ifstream cf;

	static SCRKEYWORDS interp[] = {
		{"slog", &ScriptInterp::scrSlog, &ScriptCommand::chkHasArgs},
		{"exit", &ScriptInterp::scrExit, &ScriptCommand::chkNoArgs},
		{"set", &ScriptInterp::scrSet, &ScriptCommand::chkIgnore},
		{"clear", &ScriptInterp::scrClear, &ScriptCommand::chkHasVars},
		{"try", &ScriptInterp::scrTry, &ScriptCommand::chkHasArgs},
		{"goto", &ScriptInterp::scrGoto, &ScriptCommand::chkHasArgs},
		{"call", &ScriptInterp::scrCall, &ScriptCommand::chkHasArgs},
		{"gosub", &ScriptInterp::scrCall, &ScriptCommand::chkHasArgs},
		{"pop", &ScriptInterp::scrPop, &ScriptCommand::chkIgnore},
		{"return", &ScriptInterp::scrReturn, &ScriptCommand::chkNoArgs},
		{"select", &ScriptInterp::scrSelect, &ScriptCommand::chkHasArgs},
		{"size", &ScriptInterp::scrSize, &ScriptCommand::chkHasArgs},
		{"init", &ScriptInterp::scrInit, &ScriptCommand::chkIgnore},
		{"if", &ScriptInterp::scrIf, &ScriptCommand::chkHasArgs},
		{"do", &ScriptInterp::scrDo, &ScriptCommand::chkIgnore},
		{"for", &ScriptInterp::scrFor, &ScriptCommand::chkHasList},
		{"read", &ScriptInterp::scrRead, &ScriptCommand::chkHasArgs},
		{"tryeach", &ScriptInterp::scrTryeach, &ScriptCommand::chkHasArgs},
		{"foreach", &ScriptInterp::scrForeach, &ScriptCommand::chkHasArgs},
		{"repeat", &ScriptInterp::scrRepeat, &ScriptCommand::chkHasArgs},
		{"gather", &ScriptInterp::scrGather, &ScriptCommand::chkHasArgs},
		{"loop", &ScriptInterp::scrLoop, &ScriptCommand::chkIgnore},
		{"fifo", &ScriptInterp::scrFifo, &ScriptCommand::chkHasArgs},
		{"post", &ScriptInterp::scrPost, &ScriptCommand::chkHasArgs},
		{"break", &ScriptInterp::scrBreak, &ScriptCommand::chkIgnore},
		{"continue", &ScriptInterp::scrContinue, &ScriptCommand::chkIgnore},
		{"swap", &ScriptInterp::scrSwap, &ScriptCommand::chkHasArgs},
		{"inc", &ScriptInterp::scrInc, &ScriptCommand::chkHasArgs},
		{"dec", &ScriptInterp::scrDec, &ScriptCommand::chkHasArgs},
		{"once", &ScriptInterp::scrOnce, &ScriptCommand::chkHasArgs},
		{"on", &ScriptInterp::scrOn, &ScriptCommand::chkHasArgs},
		{"pack", &ScriptInterp::scrPack, &ScriptCommand::chkHasList},
		{"unpack", &ScriptInterp::scrUnpack, &ScriptCommand::chkHasList},
		{"packed", &ScriptInterp::scrPacked, &ScriptCommand::chkHasArgs},
		{"lock", &ScriptInterp::scrLock, &ScriptCommand::chkIgnore},
		{"has", &ScriptInterp::scrHas, &ScriptCommand::chkHasArgs},
		{"missing", &ScriptInterp::scrMissing, &ScriptCommand::chkHasArgs},
		{"const", &ScriptInterp::scrConst, &ScriptCommand::chkIgnore},
		{"basename", &ScriptInterp::scrBasename, &ScriptCommand::chkHasArgs},
		{"dirname", &ScriptInterp::scrDirname, &ScriptCommand::chkHasArgs},
		{"fullpath", &ScriptInterp::scrFullpath, &ScriptCommand::chkHasArgs},
		{"alias", &ScriptInterp::scrAlias, &ScriptCommand::chkHasArgs},
		{"today", &ScriptInterp::scrToday, &ScriptCommand::chkHasArgs},
		{"data", &ScriptInterp::scrData, &ScriptCommand::chkHasArgs},
		{"dup", &ScriptInterp::scrDup, &ScriptCommand::chkHasArgs},
		{"array", &ScriptInterp::scrArray, &ScriptCommand::chkHasArgs},
		{"list", &ScriptInterp::scrList, &ScriptCommand::chkHasArgs},
		{"sequence", &ScriptInterp::scrSequence, &ScriptCommand::chkHasArgs},
		{"stack", &ScriptInterp::scrStack, &ScriptCommand::chkHasArgs},
		{"counter", &ScriptInterp::scrCounter, &ScriptCommand::chkHasArgs},
		{"use", &ScriptInterp::scrUse, &ScriptCommand::chkUse},
		{"skip", &ScriptInterp::scrSkip,  &ScriptCommand::chkHasArgs},
		{"arm", &ScriptInterp::scrArm,  &ScriptCommand::chkHasArgs},
		{"disarm", &ScriptInterp::scrDisarm, &ScriptCommand::chkHasArgs},
		{"enable", &ScriptInterp::scrEnable, &ScriptCommand::chkHasArgs},
		{"disable", &ScriptInterp::scrDisable, &ScriptCommand::chkHasArgs},
		{NULL, NULL, NULL}};

	int i;
	memset(&keywords, 0, sizeof(keywords));
	for(i = 0; i < TRAP_BITS; ++i)
		traps[i] = "<undefined>";

	active = NULL;
	keyword_count = 0;
	trap_count = 0;

	// add special default exit and error traps

	Trap("exit");
	Trap("error");
	Load(interp);

	// parse global definition entries

	if(!isDir(SCRIPT_MODULE_PATH))
		return;

	if(!canAccess(SCRIPT_MODULE_PATH))
		return;

	Dir dir(SCRIPT_MODULE_PATH);
	for(;;)
	{
		scrname = dir.getName();
		if(!scrname)
			return;

		ext = strchr(scrname, '.');
		if(!ext)
			continue;
		if(stricmp(ext, "cfg"))
			continue;

		snprintf(buffer, sizeof(buffer) - 1, "%s/%s",
			SCRIPT_MODULE_PATH, scrname);

		cf.open(buffer);
		if(!cf.is_open())
			continue;

		for(;;)
		{
			cf.getline(buffer, sizeof(buffer) - 1);
			if(cf.eof())
				break;
			scrname = buffer;
			while(isspace(*scrname))
				++scrname;

			if(*scrname == '#' || !*scrname)
				continue;

			ext = scrname;

			while(*ext && !isspace(*ext))
				++ext;

			*ext = 0;
			while(isspace(*ext) || *ext == '=')
				++ext;

			setValue(scrname, ext);
		}
		cf.close();
	}
}

scriptmethod_t ScriptCommand::getHandler(const char *keyword)
{
	keyword_t *key;
	char keybuf[33];
	int len = 0;
	char *kw = keybuf;

	while(len++ < 32 && *keyword && *keyword != '.')
		*(kw++) = (*keyword++);
	*kw = 0;
	keyword = keybuf;
 
	key = keywords[keyindex(keyword)];

	while(key)
	{
		if(!stricmp(key->keyword, keyword))
			return key->method;

		key = key->next;
	}
	if(ScriptModule::find(keyword))
		return &ScriptInterp::scrLoadable;
	return (scriptmethod_t)NULL;
}

char *ScriptCommand::Check(char *keyword, scriptline_t *line, ScriptImage *img)
{
	ScriptModule *mod;
	keyword_t *key;
	char keybuf[33];
	int len = 0;
	char *kw = keybuf;

	while(len++ < 32 && *keyword && *keyword != '.')
		*(kw++) = *(keyword++);

	*kw = 0;
	keyword = keybuf;
	key = keywords[keyindex(keyword)];

	while(key)
	{
		if(!stricmp(key->keyword, keyword))
			return Check(key->check, line, img);

		key = key->next;
	}
	mod = ScriptModule::find(keyword);
	if(mod)
		return mod->checkScript(line, img);
	return "unknown command";
}

unsigned ScriptCommand::getTrapId(const char *trapname)
{
	unsigned i;

	for(i = 0; i < TRAP_BITS; ++i)
	{
		if(!stricmp(traps[i], trapname))
			return i;
	}
	return 0;
}

unsigned long ScriptCommand::getTrapMask(unsigned id)
{
	return 1 << id;
}

unsigned long ScriptCommand::getTrapMask(const char *trapname)
{
	unsigned long mask = 1;
	int i;

	for(i = 0; i < TRAP_BITS; ++i)
	{
		if(!stricmp(traps[i], trapname))
			return mask;

		mask = mask << 1;
	}
	return 0;
}

void ScriptCommand::Load(SCRKEYWORDS *keydefs)
{
	int key;
	keyword_t *script;

	for(;;)
	{
		if(!keydefs->keyword)
			break;

		key = keyindex(keydefs->keyword);
		script = (keyword_t *)alloc(sizeof(keyword_t) + strlen(keydefs->keyword));
		strcpy(script->keyword, keydefs->keyword);
		script->method = keydefs->method;
		script->check = keydefs->check;
		script->next = keywords[key];
		keywords[key] = script;
		++keydefs;
	}
}

int ScriptCommand::Trap(const char *trapname)
{
	traps[trap_count++] = alloc((char *)trapname);
	return trap_count;
}

char *ScriptCommand::chkIgnore(scriptline_t *line, ScriptImage *img)
{
	return NULL;
}

char *ScriptCommand::chkUse(scriptline_t *line, ScriptImage *img)
{
	int argc = line->argc;
	char **argv = line->args;
	
	if(!argc)
		return "missing args";

	while(argc--)
	{
		if(!ScriptPackage::usePackage(*argv))
			return "package missing";
		++argv;
	}
	return NULL;
}

char *ScriptCommand::chkModule(scriptline_t *line, ScriptImage *img)
{
	ScriptModule *mod = ScriptModule::first;
	char keybuf[32];

	char *kw = line->cmd;
	int len = 0;
	while(*kw && len < sizeof(keybuf) - 1 && *kw != '.')
		keybuf[len++] = *kw;
	keybuf[len] = 0;

	while(mod)
	{
		if(!stricmp(kw, mod->cmd))
			break;
		return NULL;
	}
	return mod->checkScript(line, img);
}
	
char *ScriptCommand::chkHasModify(scriptline_t *line, ScriptImage *img)
{
	if(!line->argc)
		return "arguments missing";

	if(line->argc < 2)
		return "no values to modify";

	if(*(line->args[0]) != '%' && *(line->args[0]) != '@')
		return "invalid variable assignment";

	return NULL;
}

char *ScriptCommand::chkHasList(scriptline_t *line, ScriptImage *img)
{
	if(!line->argc)
		return "arguments missing";

	if(line->argc < 2)
		return "no values to assign";

	if(*(line->args[0]) != '%' && *(line->args[0]) !='@')
		return "invalid variable assignment";

	return NULL;
}

char *ScriptCommand::chkHasArgs(scriptline_t *line, ScriptImage *img)
{
	if(!line->argc)
		return "arguments missing";

	return NULL;
}

char *ScriptCommand::chkNoArgs(scriptline_t *line, ScriptImage *img)
{
	if(line->argc)
		return "invalid arguments";

	return NULL;
} 

char *ScriptCommand::chkHasVars(scriptline_t *line, ScriptImage *img)
{
	int index = 0;
	char ch;

	if(line->argc < 1)
		return "no arguments";

	while(index < line->argc)
	{
		ch = *(line->args[index++]);
		if(ch != '%' && ch != '@')
			return "variable argument required";
	}

	return NULL;
}

ScriptSymbol::ScriptSymbol(int size, int paging) :
SharedMemPager(paging)
{
	symsize = size;
	memset(index, 0, sizeof(index));
	trigger = NULL;
}

scriptsymbol_t *ScriptSymbol::getTrigger(void)
{
	scriptsymbol_t *sym;

	if(!trigger)
		return NULL;

	sym = trigger;
	trigger = NULL;
	return sym;
}

void ScriptSymbol::Purge(void)
{
	trigger = NULL;
	MemPager::purge();
	memset(index, 0, sizeof(index));
}

unsigned ScriptSymbol::getIndex(const char *symname)
{
	unsigned int key = 0;

	while(*symname)
		key ^= (key << 1) ^ (*(symname++) & 0x1f);		
	
	return key % SYMBOL_INDEX_SIZE;
}

void ScriptSymbol::Commit(scriptsymbol_t *sym)
{
	char array[96];
	char *id, *cp;
	int count;
	int index;

	if(sym->flags.type != SYM_TYPE_INDEX)
		return;

	strcpy(array, sym->id);
	id = strrchr(array, '.');
	if(!id)
		return;

	strcpy(id, ".limit");
	cp = getSymbol(array);
	if(!cp)
		return;

	count = atoi(cp);
	index = atoi(sym->data);
	if(index < 1 || index > count)
		index = 0;
	sprintf(sym->data, "%d", index);
	*id = 0;
	sym = getAlias(array);
	if(!sym)
		return;
	snprintf(sym->data, sym->flags.size, "%s.%d", array, index);
}

scriptsymbol_t *ScriptSymbol::getEntry(const char *symname, int size)
{
	int key;
	scriptsymbol_t *sym;

	EnterMutex();

retry:
	if(*symname == '%')
		++symname;
	key = getIndex(symname);
	sym = index[key];
	while(sym)
	{
		if(!stricmp(sym->id, symname))
			break;

		sym = sym->next;
	}

	if(sym)
	{
		if(sym->flags.type == SYM_TYPE_ALIAS)
		{
			symname = sym->data;
			goto retry;
		}
	}

	if(sym || !size)
	{
		LeaveMutex();
		return sym;
	}

	sym = (scriptsymbol_t *)alloc(sizeof(scriptsymbol_t) + size);
	sym->id = MemPager::alloc((char *)symname);
	sym->next = index[key];
	sym->flags.size = size;
	sym->flags.commit = false;
	sym->flags.initial = true;
	sym->flags.system = false;
	sym->flags.readonly = false;
	sym->flags.type = SYM_TYPE_NORMAL;
	sym->data[0] = 0;
	index[key] = sym;
	LeaveMutex();
	return sym;
}

scriptsymbol_t *ScriptSymbol::getAlias(const char *symname)
{
	int key;
	scriptsymbol_t *sym;

	EnterMutex();
	if(*symname == '%')
		++symname;
	key = getIndex(symname);
	sym = index[key];
	while(sym)
	{
		if(!stricmp(sym->id, symname))
			break;

		sym = sym->next;
	}

	if(sym)
	{
		if(sym->flags.type != SYM_TYPE_ALIAS)
			sym = NULL;
	}
	LeaveMutex();
	return sym;
}


char *ScriptSymbol::getSymbol(const char *id)
{
	scriptsymbol_t *sym = getEntry(id);
	if(!sym)
		return NULL;

	if(sym->flags.initial)
		return NULL;

	return readSymbol(sym);
}

bool ScriptSymbol::setAlias(const char *id, const char *source)
{
	scriptsymbol_t *sym; 

	if(*id == '%')
		++id;

	if(*source == '%')
		++source;

	sym = getEntry(id, strlen(source));
	if(!sym->flags.initial)
		return false;

	EnterMutex();
	strcpy(sym->data, source);
	sym->flags.initial = false;
	sym->flags.readonly = true;
	sym->flags.type = SYM_TYPE_ALIAS;
	LeaveMutex();
	return true;
}	

bool ScriptSymbol::makeCounter(const char *id)
{
	scriptsymbol_t *sym;

	if(*id == '%')
		++id;

	sym = getEntry(id, 11);
	if(!sym->flags.initial)
		return false;
	EnterMutex();
	sym->flags.initial = false;
	sym->flags.type = SYM_TYPE_COUNTER;
	strcpy(sym->data, "0");
	LeaveMutex();
	return true;
}

bool ScriptSymbol::makeFifo(const char *id, unsigned char size, unsigned char rec)
{
	scriptsymbol_t *sym;

	if(*id == '%')
		++id;

	sym = getEntry(id, size * (rec + 1) + 5);
	if(!sym->flags.initial)
		return false;

	EnterMutex();
	sym->flags.initial = false;
	sym->flags.readonly = true;
	sym->flags.type = SYM_TYPE_FIFO;
	sym->data[0] = 0;
	sym->data[1] = 0;	// head
	sym->data[2] = 0;	// tail
	sym->data[3] = rec;	// rec
	sym->data[4] = (sym->flags.size - 5) / (rec + 1);
	LeaveMutex();
	return true;
}

bool ScriptSymbol::makeStack(const char *id, unsigned char size, unsigned char rec)
{
	scriptsymbol_t *sym;

	if(*id == '%')
		++id;

	sym = getEntry(id, size * (rec + 1) + 5);
	if(!sym->flags.initial)
		return false;

	EnterMutex();
	sym->flags.initial = false;
	sym->flags.readonly = true;
	sym->flags.type = SYM_TYPE_STACK;
	sym->data[0] = 0;
	sym->data[1] = 0;	// head
	sym->data[2] = 0;	// tail
	sym->data[3] = rec;	// rec
	sym->data[4] = (sym->flags.size - 5) / (rec + 1);
	LeaveMutex();
	return true;
}

bool ScriptSymbol::makeSequence(const char *id, unsigned char size, unsigned char rec)
{
	scriptsymbol_t *sym;

	if(*id == '%')
		++id;

	sym = getEntry(id, size * (rec + 1) + 5);
	if(!sym->flags.initial)
		return false;

	EnterMutex();
	sym->flags.initial = false;
	sym->flags.readonly = true;
	sym->flags.type = SYM_TYPE_SEQUENCE;
	sym->data[0] = 0;
	sym->data[1] = 0;	// head
	sym->data[2] = 0;	// tail
	sym->data[3] = rec;	// rec
	sym->data[4] = (sym->flags.size - 5) / (rec + 1);
	LeaveMutex();
	return true;
}


bool ScriptSymbol::postSymbol(scriptsymbol_t *sym, const char *value)
{
	unsigned char head, tail, pos, size, rec;

	if(sym->flags.type != SYM_TYPE_FIFO && sym->flags.type != SYM_TYPE_SEQUENCE && sym->flags.type != SYM_TYPE_STACK)
		return false;

	EnterMutex();
	head = sym->data[1];
	pos = tail = sym->data[2];
	size = sym->data[4];
	rec = sym->data[3];

	if(++tail >= size && sym->flags.type == SYM_TYPE_FIFO)
		tail = 0;
	if(pos >= size)
	{
		LeaveMutex();
		return false;
	}
	if(tail == head)
	{
		LeaveMutex();
		return false;
	}

	strncpy(sym->data + (pos * (rec + 1)) + 5, value, rec);
	sym->data[pos * (rec + 1) + rec + 5] = 0;
	sym->data[2] = tail;
	LeaveMutex();
	return true;
}

char *ScriptSymbol::readSymbol(scriptsymbol_t *sym)
{
	long value;
	char *data;
	unsigned char head, rec, size;
	switch(sym->flags.type)
	{
	case SYM_TYPE_COUNTER:
		value = atoi(sym->data);
		sprintf(sym->data, "%ld", ++value);
		return sym->data;
	case SYM_TYPE_STACK:
		EnterMutex();
		if(!sym->data[2])
		{
			LeaveMutex();
			return sym->data;
		}
		head = --sym->data[2];
		rec = sym->data[3];
		LeaveMutex();
		return sym->data + 5 + (head * (rec + 1));
	case SYM_TYPE_SEQUENCE:
		EnterMutex();
		if(sym->data[1] == sym->data[2])
			sym->data[1] = 0;
		LeaveMutex();	
	case SYM_TYPE_FIFO:
		EnterMutex();
		if(sym->data[1] == sym->data[2])
		{
			LeaveMutex();
			return sym->data;
		}
		size = sym->data[4];
		rec = sym->data[3];
		head = sym->data[1];
		data =  sym->data + head * (rec + 1) + 5;
		if(++head >= size)
			head = 0;
		sym->data[1] = head;	
		LeaveMutex();
		return data;
	default:
		return sym->data;
	}
}
		
char *ScriptSymbol::setConst(const char *id, const char *value)
{
	scriptsymbol_t *sym;

	if(*id == '%')
		++id;

	sym = getEntry(id, strlen(value));
	if(!sym)
		return NULL;

	if(!sym->flags.initial)
		return sym->data;

	EnterMutex();
	strcpy(sym->data, value);
	sym->flags.initial = false;	
	sym->flags.readonly = true;
	sym->flags.type = SYM_TYPE_NORMAL;
	LeaveMutex();
	return sym->data;
}

char *ScriptSymbol::setSymbol(const char *id, const char *value)
{
	char array[96];
	int count;
	int index;
	char *cp;
	long val;

	scriptsymbol_t *alias;
	scriptsymbol_t *sym = getEntry(id);
	if(!sym)
		return NULL;

	if(sym->flags.readonly)
		return NULL;

	if(!value)
		value = "";

	EnterMutex();
	sym->flags.initial = false;
	if(sym->flags.type == SYM_TYPE_COUNTER)
	{
		val = atoi(value);
		sprintf(sym->data, "%ld", --val);
	}
	else
	{
		strncpy(sym->data, value, sym->flags.size);
		sym->data[sym->flags.size] = 0;
	}
	if(sym->flags.type == SYM_TYPE_TRIGGER)
	{
		sym->flags.type = SYM_TYPE_NORMAL;
		trigger = sym;
	}

	if(sym->flags.commit)
		Commit(sym);

	LeaveMutex();
	return sym->data;
}

void ScriptSymbol::clrSymbol(const char *id)
{
	scriptsymbol_t *sym = getEntry(id);

	if(!sym)
		return;

	if(sym->flags.readonly || sym->flags.system)
		return;

	EnterMutex();
	sym->flags.initial = true;
	sym->data[0] = 0;
	if(sym->flags.commit)
		Commit(sym);
	LeaveMutex();
	return;
}

char *ScriptSymbol::setSymbol(const char *id, int size)
{
	scriptsymbol_t *sym;

	if(!size)
		size = symsize;

	sym = getEntry(id, size);
	if(!sym)
		return NULL;

	return sym->data;
}

ScriptImage::ScriptImage(ScriptCommand *cmd)
{
	char buffer[256];
	char *scrname, *ext;
	cmds = cmd;
	memset(index, 0, sizeof(index));
	refcount = 0;
	scrStream = (istream *)&scrSource;

	if(!isDir(SCRIPT_MODULE_PATH))
		return;

	if(!canAccess(SCRIPT_MODULE_PATH))
		return;
	
	Dir dir(SCRIPT_MODULE_PATH);
	for(;;)
	{
		scrname = dir.getName();
		if(!scrname)
			break;
		ext = strchr(scrname, '.');
		if(!ext)
			continue;
		if(stricmp(ext, ".src"))
			continue;
		snprintf(buffer, sizeof(buffer) - 1, "%s/%s",
			SCRIPT_MODULE_PATH, scrname);
		ScriptImage::Compile(buffer);
	}	
}

void ScriptImage::Purge(void)
{
	MemPager::purge();
	memset(index, 0, sizeof(index));
	refcount = 0;
}

void ScriptImage::Commit(void)
{
	cmds->EnterMutex();
	cmds->active = this;
	cmds->LeaveMutex();
}

char *ScriptImage::getToken(void)
{
	static char temp[513];
	char *cp = temp + 1;
	char *base = temp + 1;
	char q;

	if(*bp == '=')
	{
		++bp;
		if(*bp == ' ' || *bp == '\t' || !*bp)
			return "";
		if(*bp == '\"' || *bp == '\'')
		{
			q = *(bp++);
			while(q)
			{
				switch(*bp)
				{
				case '\\':
					++bp;
					if(!*bp)
					{
						q = 0;
						break;
					}
					switch(*bp)
					{
					case 't':
						*(cp++) = '\t';
						++bp;
						break;
					case 'b':
						*(cp++) = '\b';
						++bp;
						break;
					case 'n':
						*(cp++) = '\n';
						++bp;
						break;
					default:
						*(cp++) = *(bp++);
					}
					break;
				case 0:
					q = 0;
					break;
				default:
					if(*bp == q)
					{
						++bp;
						q = 0;
					}
					else		
						*(cp++) = *(bp++);
				}
			}
			*cp = 0;
			return base;
		}
		while(*bp != ' ' && *bp != '\t' && *bp)
			*(cp++) = *(bp++);
		*cp = 0;
		return base;	
	}

	if(!quote)
		while(*bp == ' ' || *bp == '\t')
			++bp;

	if(!quote && *bp == '#')
		return NULL;

	if(!*bp)
	{
		quote = false;
		return NULL;
	}

	if(*bp == '\"' && !quote)
	{
		++bp;
		quote = true;
	}

	if(!quote)
	{
retry:
		while(*bp && !isspace(*bp) && *bp != ',' && *bp != '=')
			*(cp++) = *(bp++);

		if(*bp == '=' && cp == base)
		{
			*(cp++) = *(bp++);
			goto retry;
		}

		if(*bp == '=' && cp == base + 1 && ispunct(*base))
		{
			*(cp++) = *(bp++);
			goto retry;
		}

		*cp = 0;
		if(*bp == ',')
			++bp;
		else if(*bp == '=')
			*(--base) = *(bp);
		if(!strcmp(base, "=="))
			return ".eq.";
		if(!strcmp(base, "="))
			return "-eq";
		if(!strncmp(base, "!=", 2))
			return ".ne.";
		if(!strncmp(base, "<>", 2))
			return "-ne";
		if(!strcmp(base, "<"))
			return "-lt";
		if(!strcmp(base, "<="))
			return "-le";
		if(!strcmp(base, ">"))
			return "-gt";
		if(!strcmp(base, ">="))
			return "-ge";
		return base;
	}

	if(isalnum(*bp) || strchr("~/:,. \t\'", *bp))
	{
		while(isalnum(*bp) || strchr("~=/:,. \t\'", *bp))
			*(cp++) = *(bp++);
	}
	else while(!isspace(*bp) && *bp && *bp != '\"')
		*(cp++) = *(bp++);

	if(*bp == '\n' || !*bp || *bp == '\"')
		quote = false;

	if(*bp == '\n' || *bp == '\"')
		++bp;

	*cp = 0;
	return base;
}	

unsigned ScriptImage::Gather(const char *suffix, scriptname_t **array, unsigned max)
{
	unsigned count = 0;
	unsigned sort = 0;
	unsigned key = 0;
	scriptname_t *scr;
	char *ext;
	int ins;
	
	while(count < max && key < KEYWORD_INDEX_SIZE)
	{
		scr = index[key];
		while(scr && count < max)
		{
			ext = strstr(scr->name, "::");
			if(!ext)
			{
				scr = scr->next;
				continue;
			}
			ext += 2;
			if(stricmp(ext, suffix))
			{
				scr = scr->next;
				continue;
			}
			sort = 0;
			while(sort < count)
			{
				if(stricmp(scr->name, array[sort]->name) < 0)
					break;
				++sort;
			}
			ins = count;
			while(ins > sort)
			{
				array[ins] = array[ins - 1];
				--ins;
			}
			array[sort] = scr;
			++count;	
			scr = scr->next;
		}
		++key;
	}
	return count;
}

scriptname_t *ScriptImage::dupScript(const char *src, const char *dest)
{
	scriptname_t *scr = getScript(dest);
	scriptname_t *nscr;
	unsigned key;
	if(scr)
		return scr;

	scr = getScript(src);
	if(!scr)
		return NULL;

	if(scr->mode == scriptname_t::SCRIPT_TYPE_COPY)
		return NULL;

	key = keyindex(dest);
	++duplock;
	nscr = (scriptname_t *)alloc(sizeof(scriptname_t));
	memcpy(nscr, scr, sizeof(scriptname_t));
	nscr->name = alloc((char *)dest);
	nscr->mode = scriptname_t::SCRIPT_TYPE_COPY;
	scr->mode = scriptname_t::SCRIPT_TYPE_COPIED;
	nscr->next = index[key];
	index[key] = nscr;
	--duplock;
	return nscr;
}

scriptname_t *ScriptImage::getScript(const char *name)
{
	int key = keyindex(name);
	scriptname_t *scr = index[key];

	while(scr)
	{
		if(!stricmp(scr->name, name))
			break;

		scr = scr->next;
	}
	return scr;
}

int ScriptImage::Compile(const char *scrname)
{
	char buffer[strlen(scrname) + 1];
	char *token;
	char *ext;

	strcpy(buffer, scrname);
	token = strrchr(buffer, '/');
	if(!token)
		token = buffer;
	else
		++token;

	ext = strrchr(token, '.');
	if(ext)
		*ext = 0;

	return Compile(scrname, token);
}

scriptname_t *ScriptImage::Include(const char *scrname)
{
	char buffer[strlen(scrname) + 1];
	char *token;
	char *ext;
	scriptname_t *inc;

	strcpy(buffer, scrname);
	token = strrchr(buffer, '/');
	if(!token)
		token = buffer;
	else
		++token;

	ext = strrchr(token, '.');
	if(ext)
		*ext = 0;

	inc = getScript(token);
	if(!inc)
	{
		Compile(scrname, token);
		inc = getScript(token);
	}
	return inc;
}

int ScriptImage::Compile(const char *scrname, char *name)
{
	scrSource.open(scrname);
	if(!scrSource.is_open())
		return 0;

	Compile((istream *)&scrSource, name, scrname);
}
		
int ScriptImage::Compile(istream *str, char *name, const char *scrname)
{
	const char *basename = name;
	char namebuf[256];
	char *command, *token;
	char *args[33];
	char *err, *cp;
	bool trapflag, errflag;
	int argc, i, key;
	unsigned short count, number, total = 0;
	scriptname_t *script;
	scriptline_t *line, *last;
	unsigned long addmask, submask, trapmask, mask, cmask;
	unsigned branch;
	scriptmethod_t handler;
	unsigned char loopid, looplevel;
	scriptname_t *tmpl = NULL, *base = NULL;
	bool sub = false;
	streampos pos;

	scrStream = str;

compile:
	trapflag = false;
	count = number = 0;
	last = NULL;
	bool req = false;
	addmask = submask = trapmask = 0;
	branch = 0;
	handler = NULL;
	loopid = looplevel = 0;
	bool then = false;

	key = keyindex(name);
	script = (scriptname_t *)alloc(sizeof(scriptname_t));	
	memset(script, 0, sizeof(scriptname_t));
	script->name = alloc(name);
	script->mask = 0;
	script->next = index[key];
	script->mode = scriptname_t::SCRIPT_TYPE_ORIGINAL;
	index[key] = script;

	if(!base)
		base = script;
 
	if(sub)
	{
		sub = false;
		memcpy(script->trap, base->trap, sizeof(base->trap));
	}
	
	for(;;)
	{	
		quote = false;
		if(!then)
		{
			cmask = 0;
			errflag = false;
			scrStream->getline(buffer, sizeof(buffer) - 1);
			if(scrStream->eof())
				break;

			bp = buffer;
		}
		else
			then = false;

		++number;
		while(NULL != (token = getToken()))
		{
			if(!strncmp(token, "::", 2))
			{
				name = token + 2;
				break;
			}
			if(*token == '^')
			{
				if(!trapflag)
				{
					trapmask = 0;
					trapflag = true;
				}
			}
			if(*token != '^' && *token != '+' && *token != '-' && *token != '?')
			{
				if(*token >= '0' && *token <= '9')
				{
					branch |= 1 << (*token - '0');
				}
				else
					break;
			}

			if(*token == '^')
				mask = cmds->getTrapMask(token + 1);
			else
				mask = cmds->getTrapModifier(token + 1);

			if(!mask)
			{
				slog(SLOG_ERROR) << token + 1 << ": unknown trap id; " << script->name << "(" << number << ")" << endl;
				continue;
			}

			switch(*token)
			{
			case '^':
				last = NULL;
				script->mask |= mask | cmds->getTrapDefault();
				trapmask |= mask;
				break;
			case '+':
				addmask |= mask;
				break;
			case '-':
				submask |= mask;
				break;
			case '?':
				cmask |= mask;
			}
		}
		if(!token)
			continue;


		if(!strncmp(token, "::", 2))
			break;

		if(!stricmp(token, ".module"))
		{
			name = getToken();
			if(name)
				break;
			continue;
		}

		if(!stricmp(token, ".sub"))
		{
			name = getToken();
			if(name)
			{
				sub = true;
				break;
			}
			continue;
		}

		if(!stricmp(token, ".requires"))
		{
			while(NULL != (token = getToken()));
			{
				if(getHandler(token))
					continue;
				req = true;
				slog(SLOG_CRITICAL) << token << ": required keyword missing" << endl;
			}
			if(req)
				exit(-1);
			continue;
		}	

		if(!stricmp(token, ".include") && scrStream == (istream *)&scrSource)
		{
			token = getToken();
			if(!token)
				continue;
			
			if(*token != '/')
			{
				strcpy(namebuf, scrname);
				cp = strrchr(namebuf, '/');
				if(cp)
				{
					strcpy(cp + 1, token);
					token = namebuf;
				}
			}
			pos = scrSource.tellg();
			scrSource.close();
			Include(token);
			scrSource.open(scrname, ios::in); 
			scrSource.seekg(pos);
			continue;
		}

		if(!stricmp(token, ".template") && scrStream == (istream *)&scrSource)
		{
			token = getToken();
			if(*token != '/')
			{
				strcpy(namebuf, scrname);
				cp = strrchr(namebuf, '/');
				if(cp)
				{
					strcpy(cp + 1, token);
					token = namebuf;
				}
			}
			pos = scrSource.tellg();
			scrSource.close();
			tmpl = Include(token);
			scrSource.open(scrname, ios::in);
			scrSource.seekg(pos);
			if(!tmpl)
				continue;

			for(i = 0; i < TRAP_BITS; ++i)
			{
				if(!script->trap[i])
					script->trap[i] = tmpl->trap[i];
			}	
			continue;
		}

		if(!stricmp(token, ".define"))
		{
			while(NULL != (token = getToken()))
			{
				if(!cmds->getLast(token))
					cmds->setValue(token, getToken());
				else
					getToken();
			}
			continue;
		}

		if(*token == '!')
		{
			errflag = true;
			++token;
		}
		trapflag = false;
		handler = cmds->getHandler(token);
		if(!handler)
		{
			addmask = submask = 0;
			slog(SLOG_ERROR) << token << ": unknown keyword; " << script->name << "(" << number << ")" << endl;
			continue;
		}

		command = alloc(token);	
	
		argc = 0;
		while(argc < 32 && NULL != (token = getToken()))
		{
			if(*token == '$')
			{
				token = (char *)cmds->getLast(++token);
				if(!token)
					token = "";
			}
			else
				token = alloc(token);
			
			args[argc++] = token;
			if(!stricmp(token, "then") && !stricmp(command, "if"))
			{
				--bp;
				*bp = ' ';
				then = true;
				handler = &ScriptInterp::scrIfThen;
				break;
			}
		}

		args[argc++] = NULL;
		line = (scriptline_t *)alloc(sizeof(scriptline_t));
		line->line = number;
		line->cmask = cmask;
		line->mask = ((~0 & ~trapmask) | addmask) & ~submask;
		if(script->mask)
			line->mask &= cmds->getTrapHandler(script);
		line->next = NULL;
		line->args = (char **)alloc(sizeof(char *) * argc);
		line->argc = --argc;
		line->error = errflag;
		line->method = handler;
		line->cmd = command;
		line->loop = 0;

		addmask = submask = 0;

		if(!stricmp(command, "repeat") || !stricmp(command, "for") || !stricmp(command, "do") || !stricmp(command, "foreach"))
		{
			if(!looplevel)
				++loopid;
			++looplevel;
			line->loop = loopid * 256 + looplevel;
		}

		if(!stricmp(command, "loop"))
		{
			line->loop = loopid * 256 + looplevel;
			if(!looplevel)
			{
				slog(SLOG_ERROR) << "loop: nesting error" << script->name << "(" << number << ")" << endl;
				continue;
			}
			else
				--looplevel;
		}
		memcpy(line->args, &args, sizeof(char *) * argc);

		err = cmds->Check(command, line, this);
		if(err)
		{
			slog(SLOG_ERROR) << command << ": " << err << "; " << script->name << "(" << number << ")" << endl;
			continue;
		}

		++count;
		script->mask |= trapmask;
		if(!script->first)
			script->first = line;

		if(trapmask && !last)
		{
			for(i = 0; i < TRAP_BITS; ++i)
			{
				if((1l << i) & trapmask)
				{
					if(!script->trap[i])
						script->trap[i] = line;
				}
			}
		}
		if(branch)
		{
			for(i = 0; i < 10; ++i)
			{
				if((1l << i) & branch)
				{
					if(!script->skip[i])
						script->skip[i] = line;
				}
			}
		}

		if(last)
			last->next = line;
			
		last = line;
		branch = 0;
	}
	line = script->first;
	if(!script->mask)
		script->mask = cmds->getTrapDefault();
	while(line)
	{
		line->mask &= script->mask;
		line = line->next;
	}
	total += count;
	slog(SLOG_INFO) << script->name << ": " << count << " steps compiled" << endl;
	if(!scrStream->eof())
	{
		strcpy(namebuf, basename);
		strcat(namebuf, "::");
		strcat(namebuf, name);
		name = namebuf;
		goto compile;
	}
	scrSource.close();
	return total;
}

